/*
 * Cosmos - an xscreensaver module.
 * Tom Campbell and Jim Tanner Oct 2000 <campbell@mindspring.com>
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose and without fee is hereby granted,
 * provided that the above copyright notice appear in all copies and that
 * both that copyright notice and this permission notice appear in
 * supporting documentation.
 *
 * This file is provided AS IS with no warranties of any kind.  The author
 * shall have no liability with respect to the infringement of copyrights,
 * trade secrets or any patents by this file or any part thereof.  In no
 * event will the author be liable for any lost revenue or profits or
 * other special, indirect and consequential damages.
 *
 * Portions of this software were inspired by the late Seumas MacNally.
 * http://www.gamedev.net/community/memorial/seumas/
 *
 * actor.cpp: Class for painting pixels. Actors work together in casts, and 
 * follow "scripts". A script is not a traditional "script" per se, but rather
 * a means for selecting specific actor actions i.e. EXPLODE, FOLLOW, 
 * ROT_SPLINE, etc. Hopefully this will be extensable by subclassing Actors 
 * and/or Casts (but not yet).
 *
 * $Id: actor.cpp,v 1.2 2001/02/14 05:32:50 jtanner Exp $
 */


extern "C" {
#include "screenhack.h"
}

#include "stream.h"
#include "stdlib.h"

#include "actor.h"
#include "points.h"
#include "chasescript.h"

#define fabs(a) ((a) >= 0 ? (a) : -(a))

CActor::CActor(CCosmos *cosmos)
{
	this->cosmos = cosmos;
	m_init=0;
	m_leader=NULL;
	script=NULL;
}
	
CActor::CActor(CCosmos *cosmos,int x,int y)
{
	this->cosmos = cosmos;
	m_x = x;
	m_y = y;
	m_xdelta=0;
	m_ydelta=0;
	m_count =random()%100+900;
	m_visible=1;
	m_init=0;
	m_leader=NULL;
	script=NULL;
}

CActor::CActor(CCosmos *cosmos,int x,int y, CScript *script)
{
	this->cosmos = cosmos;
	this->script = script;
	m_x = x;
	m_y = y;
	m_xdelta=0;
	m_ydelta=0;
	m_count =random()%100+900;
	m_visible=1;
	m_init=0;
	m_leader = NULL;
}

void CActor::Script_SetBlastCount(int blastcount)
{
	this->blastcount = blastcount;
}

void CActor::SetLeader(CActor *leader)
{
	if (!leader) { // It's NULL, I must be the lead
		m_follow = 0;
		m_visible = 0;
	}
	else {
		m_leader = leader;
		m_follow = 1;
		m_visible = 1;
	}

}
	
void CActor::Reset(int x, int y,CScript *script)
{
	//cout << "Reseting Actors" << endl;
	this->script = script;
	m_TotalClicks = 0;
	m_count = 0;
	m_x = x;
	m_y = y;
}
