/*
 * Cosmos - an xscreensaver module. 
 * Tom Campbell and Jim Tanner Oct 2000 <campbell@mindspring.com>
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose and without fee is hereby granted,
 * provided that the above copyright notice appear in all copies and that
 * both that copyright notice and this permission notice appear in
 * supporting documentation.
 *
 * This file is provided AS IS with no warranties of any kind.  The author
 * shall have no liability with respect to the infringement of copyrights,
 * trade secrets or any patents by this file or any part thereof.  In no
 * event will the author be liable for any lost revenue or profits or
 * other special, indirect and consequential damages.
 *
 * Portions of this software were inspired by the late Seumas MacNally.
 * http://www.gamedev.net/community/memorial/seumas/
 *
 * cosmos.cpp The main hanlder of the module. Performs object notification 
 * and X interaction.
 *
 * $Id: cosmos.cpp,v 1.5 2001/02/20 03:14:31 jtanner Exp $
 */

extern "C" {
#include "screenhack.h"
#include <X11/Xutil.h>
#include <limits.h>

#include "stream.h"


#ifdef HAVE_XSHM_EXTENSION
# include "xshm.h"
#endif /* HAVE_XSHM_EXTENSION */

#ifdef HAVE_XPM
# include <X11/xpm.h>
# ifndef PIXEL_ALREADY_TYPEDEFED
#  define PIXEL_ALREADY_TYPEDEFED /* Sigh, Xmu/Drawing.h needs this... */
# endif
#endif

#ifdef HAVE_XMU
# ifndef VMS
#  include <X11/Xmu/Drawing.h>
# else  /* VMS */
#  include <Xmu/Drawing.h>
# endif /* VMS */
#endif /* HAVE_XMU */

}

#include "cosmos.h"
#include "chasescript.h"
#include "explodescript.h"
//#define HALFBURN
//#define WATCHFRAME

#define MAX_VAL             255
#define min(a,b)      ((a)<(b) ? (a) : (b))

CCosmos* CCosmos::m_instance = NULL;

void CCosmos::GetXInfo(Display *disp, Window win)
{
  XGetWindowAttributes(disp,win,&xwa);
  window   = win;
  display  = disp;
  colormap = xwa.colormap;
  depth    = xwa.depth;
  visual   = xwa.visual;
  width    = xwa.width;
  height   = xwa.height;
	
	//cout << "depth=" << xwa.depth << endl;
}

CCosmos::~CCosmos()
{

}

void CCosmos::InitColors(void)
{
	int h1, h2;
	double s1,s2;
	double v1,v2;

	h1 = color; s1 = 1.00; v1 = 0.00;
	h2 = color; s2 = 0.00; v2 = 1.00;
	colors = (XColor *) malloc(sizeof(*colors) * (ncolors+1));
	make_color_ramp (display, colormap, h1, s1, v1, h2, s2, v2, colors, &ncolors,
									 False, True, False);
}

void CCosmos::InitColors(unsigned int r,unsigned int g,unsigned int b)
{
	int h1, h2;
	double s1,s2;
	double v1,v2;

  rgb_to_hsv (0, 0, 0, &h1, &s1, &v1);
  rgb_to_hsv (r, g, b, &h2, &s2, &v2);
	colors = (XColor *) malloc(sizeof(*colors) * (ncolors+1));
	make_color_ramp (display, colormap, h1, s1, v1, h2, s2, v2, colors, &ncolors,
									 False, True, False);
}

void CCosmos::CycleColor()
{
	int h1, h2;
	double s1,s2;
	double v1,v2;

	int negative = 0;

	color++;
	if (color>360)
		color = 0;
	if (negative) {
		h1 = color; s1 = 0.00; v1 = 1.00;
		h2 = color; s2 = 1.00; v2 = 0.00;
	} else {
		h1 = color; s1 = 1.00; v1 = 0.00;
		h2 = color; s2 = 0.00; v2 = 1.00;
	}

	make_color_ramp (display, colormap, h1, s1, v1, h2, s2, v2, colors, &ncolors,
									 False, True, False);

}

void
CCosmos::MakeImage()
{
  XGCValues gcv;

#ifdef HAVE_XSHM_EXTENSIONS
  shared = True;
  xim = create_xshm_image (display, visual, depth, ZPixmap, NULL,
                           &shminfo, width, height);
#else  /* !HAVE_XSHM_EXTENSION */
  xim = 0;
#endif /* !HAVE_XSHM_EXTENSION */

  if (!xim)
    {
      shared = False;
      xim = XCreateImage (display, visual, depth, ZPixmap, 0, NULL,
                          width, height, 32, 0);
      if (xim)
        xim->data = (char *) calloc(xim->height, xim->bytes_per_line);
      if (!xim || !xim->data)
        {
          fprintf(stderr,"%s: out of memory.\n", progname);
          exit(1);
        }
    }

  gc = XCreateGC(display,window,0,&gcv);
  if (!gc) exit (1);
	fbuff = (unsigned int *) malloc( sizeof(int) * width * height);

	// Set the pointer-to-member to the correct Burn member funtion.
	Burn = xim->bits_per_pixel==8?&CCosmos::BurnCatchall:
		xim->bits_per_pixel==16?&CCosmos::Burn16:
		xim->bits_per_pixel==24?&CCosmos::Burn32:
		xim->bits_per_pixel==32?&CCosmos::Burn32:&CCosmos::BurnCatchall;
}



void CCosmos::DisplayImage(void)
{
	XPutImage(display, window, gc, xim, 0, 0, 0, 0, width, height);
}


CCosmos* CCosmos::GetInstance()
{
  if(m_instance == NULL)
    m_instance = new CCosmos();
  return m_instance;
}

CCosmos::CCosmos()
{
	cyclecolor=1;
	color=248;
	ncolors=256;
	showstars = get_boolean_resource("showstars","Boolean");
	showflame = get_boolean_resource("showflame","Boolean");
	sparkle = get_boolean_resource("sparkle","Boolean");
	stars =  get_integer_resource("stars","Integer");
	actorcount = get_integer_resource("actors","Integer");
	castcount = get_integer_resource("casts","Integer");
	resetcount = get_integer_resource("resetcount","Integer");
}

void CCosmos::InitDisplay(Display* dpy, Window win)
{

	GetXInfo(dpy,win);
	InitColors();
	MakeImage();
	frameclock = clock();
	CreateCasts();
	frame=0;
}

// This is the slow mulipurpose burn function. It will go away someday.
// For now, it is the catchall.
void CCosmos::BurnCatchall(int flame, int up)
{
#define DECAY 1
	unsigned int *buf = (unsigned int *)fbuff;
	int pitch = width;
	register int t;
	register unsigned int *p;
	static unsigned int *base=NULL;
	static int baseage=0;

	// Buffer pointers
 	unsigned char  *p8=NULL;
 	unsigned short *p16=NULL;
 	unsigned int   *p32=NULL;

 	flame = 1;
	
	if (!base) {
		base = (unsigned int *)malloc(sizeof(int)*width);
	}
	
	if (!baseage) 
		for (int i=0;i<width;i++)
			*(base+i) = random()%256; // This needs improvement;
	baseage++;
	if (baseage > 50)
		baseage = 0;

	if (up) {
		int offset = frame%2;
		//cout << offset << endl;
		if (showflame)
			memcpy(buf+((height-1)*width),base,sizeof(int)*width);
		p = buf+1+offset; 
		switch (xim->bits_per_pixel) {
		case 8 :
			p8  = (unsigned char*)xim->data + 1;
			break;
		case 16 :
			p16  = (unsigned short *)xim->data + 1 + offset;
			break;
		case 24 :
		case 32 :
			p32  = (unsigned int *)xim->data + 1;
			break;
			// No default case
		}

		unsigned int *pm = p + ((height-1) * width);
		while (p<pm) {
			if (*(p-1) || *(p) || *(p+1) || *(p+pitch)) {
				t = (*(p-1) + *(p) + *(p+1) + *(p+(pitch)) >> 2) - DECAY;
				*p = t < 0 ? 0 : t;
				if (*p) {
					switch (xim->bits_per_pixel) { // Put the pixel cast to the color mode
					case 8 :	
						*p8 = (unsigned char)colors[(int)*p].pixel;
						break;
					case 16 :	
						*p16 = (unsigned short)colors[(int)*p].pixel;
						break;
					case 24 :
					case 32 :
						*p32 = (unsigned int)colors[(int)*p].pixel;
						break;
					default :
						// Figure out the x y coordinate from p-buff
						XPutPixel(xim,(p-buf)%width,(p-buf)/width,colors[(int)*p].pixel);
					}
				}
			}
			p++;	
			// switch this?
			p8++;
			p16++;
			p32++;
		}
	} // End if(up) 
	else { // Must be Down
		if (showflame)
			memcpy(buf,base,sizeof(int)*width);
		p = buf + (height*width);
		switch (xim->bits_per_pixel) {
		case 8 :
			p8  = (unsigned char*)xim->data + (height*width);
			break;
		case 16 :
			p16  = (unsigned short *)xim->data + (height*width);
			break;
		case 24 :
		case 32 :
			p32  = (unsigned int *)xim->data + (height*width);
			break;
			
			// No default case
		
		}
		unsigned int *pm = buf + width + 1;
		while (p!=pm) {
			if (*(p-1) || *(p) || *(p+1) || *(p-pitch)) {
				t = ((*(p-1) + *(p) + *(p+1) + *(p-width)) >> 2) - 1;
				*p = t < 0 ? 0 : t;
				if(*p) {
					switch (xim->bits_per_pixel) { // Put the pixel cast to the color mode
					case 8 :	
						*p8 = (unsigned char)colors[(int)*p].pixel;
						break;
					case 16 :	
						*p16 = (unsigned short)colors[(int)*p].pixel;
						break;
					case 24 :
					case 32 :
						*p32 = (unsigned int)colors[(int)*p].pixel;
						break;
					default :

						XPutPixel(xim,(p-buf)%width,(p-buf)/width,colors[(int)*p].pixel);
					}
				}
			}
			p--;
			p8--;
			p16--;
			p32--;
		} 
	} // End Down
}

void CCosmos::Burn16(int flame, int up)
{
#define DECAY 1
	unsigned int *buf = (unsigned int *)fbuff;
	int pitch = width;
	register int t;
	register unsigned int *p;
	static unsigned int *base=NULL;
	static int baseage=0;

	// Buffer pointers
 	unsigned short *p16;
 	flame = 1;
	//up = 0;
	if (!base) {
		base = (unsigned int *)malloc(sizeof(int)*width);
	}
	
	if (!baseage) 
		for (int i=0;i<=width;i++)
			*(base+i) = random()%256; // This needs improvement;
	baseage++;
	if (baseage > 50)
		baseage = 0;

	if (up) {
		if (showflame)
			memcpy(buf+((height-1)*width),base,sizeof(int)*width);
		p = buf+1;
		p16  = (unsigned short *)xim->data + 1;
		unsigned int *pm = p + ((height-1) * width);
		while (p<pm) {
			if ((t=(*(p-1) + *(p) + *(p+1) + *(p+pitch)))) {
				t = (t  >> 2) - DECAY;
				*p = t < 0 ? 0 : t;
				if (*p) {

					*p16 = (unsigned short)colors[(int)*p].pixel;
				}
			}
			p++;	
			p16++;
		}
	} // End if(up) 
	else { // Must be Down
		if (showflame)
			memcpy(buf,base,sizeof(int)*width);
		p = buf + (height*width);
		p16  = (unsigned short *)xim->data + (height*width);
		unsigned int *pm = buf + width + 1;
		while (p>pm) {
			if ((t=(*(p-1) + *(p) + *(p+1) + *(p-pitch)))) {
				t = (t >> 2) - 1;
				*p = t < 0 ? 0 : t;
				if(*p) {
						*p16 = (unsigned short)colors[(int)*p].pixel;
				}
			}
			p--;
			p16--;
		} 
	} // End Down
}

void CCosmos::Burn32(int flame, int up)
{
#define DECAY 1
	unsigned int *buf = (unsigned int *)fbuff;
	int pitch = width;
	register int t;
	register unsigned int *p;
	static unsigned int *base=NULL;
	static int baseage=0;


 	unsigned int *p32;
 	flame = 1;
	if (!base) {
		base = (unsigned int *)malloc(sizeof(int)*width);
	}
	
	if (!baseage) 
		for (int i=0;i<width;i++)
			*(base+i) = random()%256; // This needs improvement;
	baseage++;
	if (baseage > 50)
		baseage = 0;

	if (up) {
		if (showflame)
			memcpy(buf+((height-1)*width),base,sizeof(int)*width);
		p = buf+1; 
		p32  = (unsigned int *)xim->data + 1;
		unsigned int *pm = p + ((height-1) * width);
		while (p<pm) {
			if ((t=(*(p-1) + *(p) + *(p+1) + *(p+pitch)))) {
				t = (t  >> 2) - DECAY;
				*p = t < 0 ? 0 : t;
				if (*p) {
					*p32 = (unsigned int)colors[(int)*p].pixel;
				}
			}
			p++;	
			p32++;
		}
	} // End if(up) 
	else { // Must be Down
		if (showflame)
			memcpy(buf,base,sizeof(int)*width);
		p = buf + (height*width);
		p32  = (unsigned int *)xim->data + (height*width);
		unsigned int *pm = buf + width + 1;
		while (p!=pm) {
			if ((t=(*(p-1) + *(p) + *(p+1) + *(p-pitch)))) {
				t = (t >> 2) - 1;
				*p = t < 0 ? 0 : t;
				if(*p) {
						*p32 = (unsigned int)colors[(int)*p].pixel;
				}
			}
			p--;
			p32--;
		} 
	} // End Down
}

void CCosmos::Sparkle()
{
	int x,y;
	unsigned int *data = (unsigned int *) fbuff;
	unsigned int *p;
	int pitch = width;

	for(y = 0; y < height; y++){
		p = (unsigned int*)data + y * pitch + 1;
		x = 1 + random() % width -2;
		if (p[x] > 50 && p[x] < 235 ) 
			p[x] = 250;
	}
}
void CCosmos::Stars()
{
	static int init=0;
	static int *x=NULL;
	static int *y=NULL;
	static int age=0;
	// BUG BUG - We can't free this. We must depend on the OS to do it at
	// termination.
	if (!x)
		x = (int*) malloc(sizeof(int) * stars);
	if (!y)
		y = (int*) malloc(sizeof(int) * stars);

	if (50==age) {
		for (int i=0;i<stars;i++) {
			x[i]--;
			if (x[i]<0)
				x[i]=width;
		}
		age=0;
	}
	age++;
	
	if (!init) {
		for (int i=0;i<stars;i++) {
			x[i] = random()%width;
			y[i] = random()%(height-10);
		}
		init =1;
	}
	
	for(int i = 0;i<stars;i++) {
		SetPixel(x[i],y[i],255);
	}
}

void CCosmos::CreateCasts()
{
	CScript *script;

	for(int i=0;i<castcount;i++) {
		int scriptnum = random()%SCRIPTS;
		switch (scriptnum) {
		case EXPLODE :
			script = new CExplodeScript();
			break;
		case CHASE :
			script = new CChaseScript();
			break;
		default: // JIC
		  cout << "CreateCasts():  Something's wrong--we're at the default case" << endl;
			script = new CExplodeScript();
			
		}
		
		cast.push_back(new CCast(this,script));
	}
}

void CCosmos::ResetCasts()
{
	/* TODO Get Reset Working */

	for(int i=0;i<castcount;i++) {
		delete(cast[i]->script);
		CScript *script;

		int scriptnum = random()%SCRIPTS;
		switch (scriptnum) {
		case EXPLODE :
			script = new CExplodeScript();
			break;
		case CHASE :
			script = new CChaseScript();
			break;
		default: // JIC
		  cout << "ResetCasts():  Something's wrong--we're at the default case" << endl;
			script = new CExplodeScript();

		}	
		cast[i]->script = script;
		cast[i]->Reset(script);
 	}
}
void CCosmos::UpdateCasts()
{
	for(int i=0;i<castcount;i++) {
		cast[i]->Update();
	}
}


// A few primatives we need
//
void CCosmos::SetPixel(int x,int y,int c)
{
	// Hopefully at some point it can be assumed the pixels are sane. but for now drop them.
	if (x >= width || y >= height || x < 0 || y < 0) {
		//cout << "Ignoring out of bounds pixel : x=" << x << " y=" << y << endl;
		return;
	}

	unsigned int *p = fbuff +x +(y*width);
	*p = c;
}
// DrawLine
// Using Bresenham's algorithm.. could be improved
//
void CCosmos::DrawLine(int x0, int y0, int x1, int y1, int color)
{
	
	int dy = y1 - y0;
	int dx = x1 - x0;
	int stepx, stepy;

	if (dy < 0) { dy = -dy;  stepy = -1; } else { stepy = 1; }
	if (dx < 0) { dx = -dx;  stepx = -1; } else { stepx = 1; }
	dy <<= 1;                        
	dx <<= 1;                        

	SetPixel( x0, y0,color);
	if (dx > dy) {
		int fraction = dy - (dx >> 1); 
		while (x0 != x1) {
			if (fraction >= 0) {
				y0 += stepy;
				fraction -= dx;  
			}
			x0 += stepx;
			fraction += dy;    
			SetPixel(x0, y0,color);
		}
	} else {
		int fraction = dx - (dy >> 1);
		while (y0 != y1) {
			if (fraction >= 0) {
				x0 += stepx;
				fraction -= dy;
			}
			y0 += stepy;
			fraction += dx;
			SetPixel( x0, y0,color);
		}
	}
}

void CCosmos::Run()
{
	static int up=0;
	static int flame=1;
	
	if (!(frame%resetcount)) {
		ResetCasts();
	} 
	if (!(frame%1700))
		up = !up;
	if (showstars)
		Stars();
	if (sparkle)
		Sparkle();
	if (cyclecolor && !(frame%25))
		CycleColor();
		
	UpdateCasts();
	(this->*Burn)(flame,up);
	DisplayImage();
	frame++;
	
#ifdef WATCHFRAME
  // Draw a simple frame counter.
	static unsigned long framecount=0;
	if (!(frame%100))
		cout << "Frame=" << frame << endl;
	framecount++;
	if (clock() - frameclock >= CLOCKS_PER_SEC) {
		cout << "Frames:" << framecount << endl;
		framecount = 0;
		frameclock = clock();
	}
#endif
} 



char *progclass = "Cosmos";

char *defaults [] = {
  ".background:     black",
  ".foreground:     #FFAF5F",
  "*bitmap:         (default)",
  "*bitmapBaseline: 20",
  "*delay:          10000",
  "*hspread:        30",
  "*vspread:        97",
  "*residual:       99",
  "*variance:       50",
  "*vartrend:       20",
  "*bloom:          True",   

#ifdef HAVE_XSHM_EXTENSION
	"*useSHM: False",   /* xshm turns out not to help. */
#endif /* HAVE_XSHM_EXTENSION */
	".showstars:      True", // Show the stars?
	".showflame:      True", // Show the flame?
	".sparkle:        True", // Sparkle effects?
	".reset:          2000", // Reset the actors every x frames
	".changedircount: 1700", // Change the burn direction every x frames
	".changedir:      True", // Should I change the burn direction?
	".stars:          200",  // How Many stars
	".casts:            3", // How many casts?
	".actors:         100",  // How many Actors?
	".actormaxsepeed: 20",   // Max speed of the actors (in pixels per frame)
	".resetcount:     500",  // Number of frames to reset the casts
  0
};

XrmOptionDescRec options [] = {
  { "-delay",     ".delay",          XrmoptionSepArg, 0 },
  { "-bitmap",    ".bitmap",         XrmoptionSepArg, 0 },
  { "-baseline",  ".bitmapBaseline", XrmoptionSepArg, 0 },
  { "-hspread",   ".hspread",        XrmoptionSepArg, 0 },
  { "-vspread",   ".vspread",        XrmoptionSepArg, 0 },
  { "-residual",  ".residual",       XrmoptionSepArg, 0 },
  { "-variance",  ".variance",       XrmoptionSepArg, 0 },
  { "-vartrend",  ".vartrend",       XrmoptionSepArg, 0 },
  { "-bloom",     ".bloom",          XrmoptionNoArg, "True" },
  { "-no-bloom",  ".bloom",          XrmoptionNoArg, "False" },
#ifdef HAVE_XSHM_EXTENSION
  { "-shm",       ".useSHM",         XrmoptionNoArg, "True" },
  { "-no-shm",    ".useSHM",         XrmoptionNoArg, "False" },
#endif /* HAVE_XSHM_EXTENSION */
	{ "-showstars", ".showstars",      XrmoptionSepArg, 0 },
	{ "-showflame", ".showflame",      XrmoptionSepArg, 0 },
	{ "-sparkle",   ".sparkle",        XrmoptionSepArg, 0},
	{ "-stars",     ".stars",          XrmoptionSepArg, 0},
	{ "-actors",    ".actors",         XrmoptionSepArg, 0},
	{ "-casts",     ".casts",          XrmoptionSepArg, 0},
	{ "-resetcount",".resetcount",     XrmoptionSepArg, 0},    
	{ 0, 0, XrmoptionNoArg, 0 }
};

void
screenhack (Display *disp, Window win)
{
	CCosmos::GetInstance()->InitDisplay(disp,win);
  while (1) {
		CCosmos::GetInstance()->Run();
		XSync(disp,False);
		screenhack_handle_events(disp);
		//if (delay)
		//	usleep (delay);
	}
}
